private import common;

private import texture;
private import fpstimer;
private import vector;
private import sound;
private import letter;

private import SDL;
private import SDL_mixer;

private import opengl;
private import openglu;

private import std.ctype;
private import std.file;

Texture tex;
Sound snd;

Rc rc;

abstract class Pbg {
    static Vec3 mpos;
    float r, g, b;
    Vec3 pos;
    Vec3 vel;
    float life;
    float alpha;

    // from NeHe19
    static float colors[12][3]=           // Rainbow Of Colors
        [
            [1.0f,0.5f,0.5f],[0.75f,1.0f,0.5f],[0.5f,0.5f,1.0f],[1.0f,0.75f,0.5f],
            [0.5f,1.0f,0.5f],[0.5f,1.0f,0.75f],[0.5f,1.0f,1.0f],[0.5f,0.75f,1.0f],
            [1.0f,1.0f,0.5f],[0.75f,0.5f,1.0f],[1.0f,0.5f,1.0f],[1.0f,0.5f,0.75f]
            ];

    this() {
        life = 1;
        alpha = 0.7;
    }
    this(Vec3 p0) {
        this();
        pos = p0;
    }
    this(Vec3 p0, Vec3 v0) {
        this(p0);
        vel = v0;
    }
    this(float r0, float g0, float b0) {
        r = r0;
        g = g0;
        b = b0;
    }
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        this(p0, v0);
        r = r0;
        g = g0;
        b = b0;
    }

    void proc() {
        pos += vel;
    }

    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, alpha);

        GLfloat[] texcoord = tex.texcoord;

        float ss = rc.getFloat("shotSize");
        float st = rc.getFloat("shotTail");
        float tx = ss + ((vx > 0) ? vx : -vx) * st;
        float ty = ss + ((vy > 0) ? vy : -vy) * st;

        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }

    float x() { return pos.x; }
    float y() { return pos.y; }
    float z() { return pos.z; }
    float x(float v) { return pos.x = v; }
    float y(float v) { return pos.y = v; }
    float z(float v) { return pos.z = v; }
    float vx() { return vel.x; }
    float vy() { return vel.y; }
    float vz() { return vel.z; }
    float vx(float v) { return vel.x = v; }
    float vy(float v) { return vel.y = v; }
    float vz(float v) { return vel.z = v; }
}

class A : Pbg {
    int n;
    int turn;
    this(Vec3 p0, int n0) {
        super(p0, Vec(0, -1, 0), 1, 1, 1);
        n = n0;
    }

    void proc() {
        super.proc();
        if (y < 480) {
            life = 0;
            Main.obj.boxPower += 20;
            Main.obj.pbgs ~= new B(pos, n);
            snd.play(n);
        }
        turn++;
        alpha = (turn > 70) ? 0.7 : turn * 0.01;
    }
}

class B : Pbg {
    Vec3 acc;
    int n;

    this(Vec3 p0, int n0) {
        n = n0;
        float[3] c = colors[(rc.getInt("enableRandomColor")) ? rnd(12) : n];
        r = c[0];
        g = c[1];
        b = c[2];
        pos = p0;
        vel = Vec(0, -3, 0);
        float accx = rc.getFloat("accelX");
        float accy = rc.getFloat("accelY");
        acc = Vec(rnd(accx*2)-accx, -rnd(accy), 0);
    }

    void proc() {
        pos += vel;
        vel += acc;
        if (y < 0) {
            if (rc.getInt("enableBound")) {
                if (rc.getInt("enableBoundSound")) {
                    snd.play(n);
                }
                if (rc.getInt("enableBoundNoAccel")) {
                    acc = Vec(0, 0, 0);
                }
                vy = -vy;
                y = vy;
            }
            else {
                life = 0;
            }
        }
        if (y > 600 || x < 0 || x > 800) {
            life = 0;
        }
    }
}

class Main {
    static Main obj_;
    static Main obj() { return obj_; }

    this() {
        obj_ = this;
    }

    Pbg[] pbgs;

    int screenWidth_, screenHeight_;

    FpsTimer timer;

    int boxPower;

	void initSDL() {
		Log.info("init SDL");

		int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO|SDL_INIT_NOPARACHUTE);
		if (ret != 0) Log.errorSDL("cannot init SDL");

		ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (ret != 0) {
			Log.warnSDL("cannot init joystick");
		}

		SDL_VideoInfo* info = SDL_GetVideoInfo();

        int bpp = rc.getInt("screenBpp");
        char[] bppMsg;

		if (bpp == -1) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using the \"best\" bpp: " ~ tostr(bpp));
		}
		else if (bpp == 0) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using current bpp");
		}
		else {
			Log.info("using user set bpp: " ~ tostr(bpp));
		}

		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
		SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

        if (rc.getInt("motionBlurEnable")) {
            SDL_GL_SetAttribute( SDL_GL_ACCUM_RED_SIZE, 5 );
            SDL_GL_SetAttribute( SDL_GL_ACCUM_GREEN_SIZE, 5 );
            SDL_GL_SetAttribute( SDL_GL_ACCUM_BLUE_SIZE, 5 );
        }

		screenWidth_ = 800;
		screenHeight_ = 600;
		int screenFull = rc.getInt("screenFull");

		int flags = SDL_OPENGL;

		if (screenFull) {
			flags |= SDL_FULLSCREEN;
		}

        Log.info("set video mode " ~
                 "size = " ~ tostr(screenWidth_) ~
                 "x" ~ tostr(screenHeight_) ~
                 ", bpp = " ~ tostr(bpp) ~
                 ", flags = " ~ tostr(flags));

		if (SDL_SetVideoMode(screenWidth_, screenHeight_,
							 bpp, flags) == null)
		{
			Log.errorSDL("cannot create screen");
		}

		info = SDL_GetVideoInfo();
		Log.info("set bpp: " ~ tostr(cast(int)info.vfmt.BitsPerPixel));

        SDL_WM_SetCaption(`pbg`, null);

        snd = new Sound();
	}

    void initGL() {
        float bgr = rc.getFloat("clearColR");
        float bgg = rc.getFloat("clearColG");
        float bgb = rc.getFloat("clearColB");
        glClearColor(bgr, bgg, bgb, 0);

        glViewport(0, 0, screenWidth_, screenHeight_);
        glLoadIdentity();
        glOrtho(-0.5, screenWidth_ - 0.5, screenHeight_ - 0.5, -0.5, -1.0, 1.0);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
//		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glBlendFunc(GL_ONE, GL_ZERO);

        // from PARSEC47
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);
        glDisable(GL_DEPTH);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);
        glLineWidth(2);
    }

    int main(char[][] args) {
        Log.init();

        rc = new Rc();
        rc.addResourceFile(`itpiano.rc`);
        initSDL();
        initGL();
        initLetter();

    all_loop:
        while (true) {
            bool running = true;
            if (args.length == 1) {
                char max = 'A';
                char[] get_rcfilename(char c) {
                    char[] f = `itpiano_`;
                    f ~= cast(char)(toupper(c));
                    f ~= `.rc`;
                    return f;
                }
                while (std.file.exists(get_rcfilename(max))) {
                    max++;
                }
                char now = '\0';

            select_loop:
                while (true) {
                    SDL_Event ev;
                    while (SDL_PollEvent(&ev)) {
                        switch (ev.type) {
                        case SDL_MOUSEMOTION:
                            int x = (ev.motion.x-50) / 30;
                            int y = ev.motion.y;
                            now = (y >= 200 && y < 230 &&
                                   x >= 0 && x < (max-'A'))
                                ? 'A'+x : '\0';
                            break;
                        case SDL_MOUSEBUTTONDOWN:
                            if (now >= 'A' && now < max) {
                                rc.addResourceFile(get_rcfilename(now));
                                break select_loop;
                            }
                            break;
                        case SDL_KEYDOWN:
                            SDLKey k = toupper(ev.key.keysym.sym);
                            if (k >= 'A' && k < max) {
                                rc.addResourceFile(get_rcfilename(k));
                                break select_loop;
                            }
                            if (k != SDLK_ESCAPE) break;
                        case SDL_QUIT:
                            running = false;
                            break all_loop;
                        default:
                        }
                    }

                    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
                    for (char c = 'A'; c < max; c++) {
                        if (c == now) {
                            drawString(""~c, 50+30*(c-'A'), 215,
                                       20, 20, 1,0,0);
                        }
                        else {
                            drawString(""~c, 50+30*(c-'A'), 215,
                                       20, 20, 1,1,1);
                        }
                    }
                    SDL_GL_SwapBuffers();
                }
            }
            else {
                rc.addResourceFile(args[1]);
            }

            tex = new Texture(`particle.bmp`);

            int enableSystemInfo = rc.getInt("enableSystemInfo");

            timer = new FpsTimer(true);
            int frameSkip = rc.getInt("frameSkip");
//        if (args.length >= 2) frameSkip = 0;
            if (frameSkip) {
                int fps = rc.getInt("fps");
                timer.setFps(fps);
            }
            else {
                timer.setFrameSkip(false);
            }

            int motionBlurEnable = rc.getInt("motionBlurEnable");
            int motionBlurFrames = rc.getInt("motionBlurFrames");
            float motionBlurAccum = rc.getFloat("motionBlurAccum");
            float motionBlurMult = rc.getFloat("motionBlurMult");
            if (motionBlurEnable && frameSkip) {
                Log.warn("motion blur enabled. so force disables frame skip.");
                frameSkip = false;
                timer.setFrameSkip(false);
            }

            pbgs = null;

            Uint32 startTicks = SDL_GetTicks();
            timer.start();
            int turn = 0;

            while (running) {
                SDL_Event ev;
                bool added = false;
                while (SDL_PollEvent(&ev)) {
                    switch (ev.type) {
                    case SDL_KEYDOWN:
                        if (ev.key.keysym.sym != SDLK_ESCAPE) break;
                        running = false;
                        break;
                    case SDL_QUIT:
                        break all_loop;
                    case SDL_MOUSEMOTION:
                        int x = (ev.motion.x-350) * 2 / 25;
                        int y = (ev.motion.y-480) / 10;
                        static int px = -1;
                        static int py = -1;
                        if (x >= 0 && x < 8 && y >= 0 && y < 10 &&
                            (!rc.getInt("enableGenerateLimit") ||
                             x != px || y != py) && !added)
                        {
                            int vx = x * 25 / 2 + 350 + 6;
                            int vy = y * 10 + 480 + 5;
                            pbgs ~= new A(Vec(vx, vy, 0), x);
                            added = true;
                            px = x;
                            py = y;
                        }
                        break;
                    default:
                    }
                }

                foreach (Pbg p; pbgs) {
                    p.proc();
                }

                Pbg[] nextpbgs;
                foreach (Pbg p; pbgs) {
                    if (p.life > 0.0) nextpbgs ~= p;
                }
                pbgs = nextpbgs;

                if (!timer.shouldSkip()) {
                    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

                    foreach (Pbg p; pbgs) {
                        p.draw();
                    }
                    if (enableSystemInfo) drawSystemInfo();

                    if (boxPower > 50) boxPower = 50;
                    else if (boxPower > 0) boxPower -= 5;
                    float c = 0.3 + boxPower * 0.01;
                    glColor4f(c, c, c, 0.8);
                    glDisable(GL_TEXTURE_2D);
                    glBegin(GL_LINE_STRIP);
                    glVertex2d(350, 480);
                    glVertex2d(350, 580);
                    glVertex2d(450, 580);
                    glVertex2d(450, 480);
                    glVertex2d(350, 480);
                    glEnd();

                    if (motionBlurEnable) {
                        glAccum(GL_ACCUM, motionBlurAccum);
                        int t = turn % motionBlurFrames;
                        if (t == motionBlurFrames-1) {
                            glAccum(GL_RETURN, 1.0);
                            SDL_GL_SwapBuffers();
                            glClear(GL_ACCUM_BUFFER_BIT);
                        }
                        else {
                            glAccum(GL_MULT, motionBlurMult);
                        }
                    }
                    else {
                        SDL_GL_SwapBuffers();
                    }
                }
                turn++;
                timer.wait();
            }

            Uint32 endTicks = SDL_GetTicks();
            printf("%d %d %d\n", turn, startTicks, endTicks);
            printf("%d FPS.\n", turn*1000/(endTicks-startTicks));
        }

        SDL_Quit();

        return 0;
    }

    void drawSystemInfo() {
        int x = 10;
        int y = 20;
        glDisable(GL_TEXTURE_2D);
        drawString(`FPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTrueFps(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`FRAMES`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTurn(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`DROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getDroppedTurn(), x+65,y, 10,12, 1,1,1);
/*
  y += 15;
  drawString(`SDROPS`, x,y, 10,12, 1,1,1);
  drawNum(timer.getSystemDroppedTurn(), x+65,y, 10,12, 1,1,1);
*/
        y += 15;
        drawString(`OBJS`, x,y, 10,12, 1,1,1);
        drawNum(pbgs.length, x+65,y, 10,12, 1,1,1);
    }
}

int main(char[][] args) {
    Main m = new Main();
    return m.main(args);
}
